library(INLA)

#Example 1: Nottingham monthly temperatures 1920-1939. ####
?nottem
data(nottem)

y=as.numeric(nottem)
plot(y,type="l")
#prepare data frame for estimation with INLA:
#(here, we handle the intercept "explicitly" as a covariate that is constant 1. we could also simply use the default internal intercept of inla(...).)
mydata=data.frame(y=as.numeric(nottem),month=rep(1:12,20),intercept=1)

#Model 1: monthly model ####
#for month, we use a cyclic random walk over the 12 months, constrained to sum to 0:
#(-1 to remove default intercept, "+intercept" to add our intercept covariate)
myformula=y~-1+intercept+f(month,model="rw1",cyclic=TRUE,constr=TRUE)

fit=inla(myformula,family="gaussian",data=mydata)
summary(fit)
#estimates of monthly mean temperatures:
fit$summary.random$month
#plot monthly mean estimates and credible intervals:
ylim=range(c(fit$summary.random$month$mean,fit$summary.random$month$`0.025quant`,fit$summary.random$month$`0.975quant`))
plot(1:12,fit$summary.random$month$mean,pch=19,ylim=ylim,xlab="month",ylab="mean temperature")
for(i in 1:12){
  lines(c(i,i),c(fit$summary.random$month$`0.025quant`[i],fit$summary.random$month$`0.975quant`[i]),col="blue",lwd=2)
}

#a look at residuals:
#to get fitted values, we have to add compute=TRUE in control.predictor:
fit=inla(myformula,
         family="gaussian",
         data=mydata,
         control.predictor=list(compute=TRUE))
resid=y-fit$summary.fitted.values$mean
#histogram of residuals (probability histogram):
hist(resid,breaks=20,freq=F)
#the model has estimated the following precision (=1/variance):
fit$summary.hyper$mean[1]
#therefore, a point estimate of the Gaussian standard deviation could be defined as follows:
sd.likelihood=1/sqrt(fit$summary.hyper$mean[1])
xvals=-10+1:1000/50
lines(xvals,dnorm(xvals,sd=sd.likelihood),col="blue",lwd=2)
#this looks OK: the Gaussian likelihood is a good model here


#Example 2: daily airquality for New York, May to September 1973.####

?airquality
plot(1:nrow(airquality),airquality$Ozone,type="l",xlab="days",ylab="ozone",lwd=2)
#some data are missing:
sum(is.na(airquality$Ozone))
#We will try to predict them with INLA.

#There are covariates (Wind, Temperature) without missing data.
#=>We can use them for prediction.

#Model 2a: linear with model with wind and temperature as covariate. ####

#Create dataframe for inla.
#To stabilize estimation, we scale the covariates to have mean 0 and variance 1:
mydata=data.frame(y=airquality$Ozone,
                  intercept=1,
                  temp=scale(airquality$Temp),
                  wind=scale(airquality$Wind)
                  )

#Define the model formula:
myformula=y~-1+intercept+temp+wind

#We set "compute=TRUE" in control.predictor to directly obtain fitted values for the missing Ozone data.
#Since the intercept estimate may be quite large, we fix a precision for fixed effects that can be chosen to be higher than the default value.
#(here, prec=0.0001 means that prior standard deviation =1/sqrt(0.0001)=100)
fit=inla(myformula,
         family="gaussian",
         data=mydata,
         control.predictor=list(compute=TRUE),
         control.fixed=list(prec=.0001))
fit2a=fit
summary(fit)
#Fixed effects: intercept, temperature and wind are significant since sign(0.025quant)==sign(0.975quant).
#Standard deviation of Gaussian likelihood is approximately as follows:
sd.likelihood=sqrt(1/fit$summary.hyperpar$mean[1])
sd.likelihood
#Show residuals of the model:
resid=airquality$Ozone-fit$summary.fitted.values$mean
#histogram of residuals (probability histogram):
hist(resid,freq=F) #still a lot of uncertainty left

#Plot prediction for missing values (blue) and fitted values (grey):
xvals=1:nrow(airquality)
plot(xvals,airquality$Ozone,type="l",xlab="days",ylab="ozone",lwd=2,ylim=c(0,200))
idx.missing=which(is.na(airquality$Ozone))
mycolor=ifelse(is.na(airquality$Ozone),"blue","gray70")
mycex=ifelse(is.na(airquality$Ozone),1,.5)
points(xvals,fit$summary.fitted.values$mean,col=mycolor,pch=19,cex=mycex)

#Can we do better for prediction?
#Are the covariate effects nonlinear?
plot(airquality$Wind,airquality$Ozone,cex=.5,pch=19)
plot(airquality$Temp,airquality$Ozone,cex=.5,pch=19)
#There could be some mild nonlinearity...



#Model 2b: First-order random walks for nonlinear covariate influence of Wind, Temperature. ####
#For each covariate, we construct a random walk over 10 levels of the covariate values.
#The estimated random walk values are constant within each of the 10 levels.

level.wind=as.integer(cut(airquality$Wind,breaks=quantile(airquality$Wind,0:10/10),include.lowest=TRUE))
#level.wind
level.temp=as.integer(cut(airquality$Temp,breaks=quantile(airquality$Temp,0:10/10),include.lowest=TRUE))
#level.temp

mydata=data.frame(y=airquality$Ozone,
                  intercept=1,
                  level.temp=level.temp,
                  level.wind=level.wind
                  )
#Penalized complexity prior for the precision:
#(we here indicate that the probability of having standard deviation larger than 10 is 0.1)
hyper.rw=list(theta=list(prior="pc.prec", initial=log(1),param=c(10,.1)))
myformula=y~-1+intercept+f(level.temp,model="rw1",scale.model=TRUE,hyper=hyper.rw)+f(level.wind,scale.model=TRUE,model="rw1",hyper=hyper.rw)

#run INLA:
fit=inla(myformula,
         family="gaussian",
         data=mydata,
         control.predictor=list(compute=TRUE),
         control.fixed=list(prec=.0001))
fit2b=fit
summary(fit)
#plot estimated nonlinear effect of temperature levels
tmp=fit$summary.random$level.temp[,c("mean","0.025quant","0.975quant")]
ylim=range(tmp)
plot(1:10,tmp$mean,type="l",ylim=ylim,lwd=2)
lines(1:10,tmp$`0.025quant`,col="blue",lwd=2)
lines(1:10,tmp$`0.975quant`,col="blue",lwd=2)
##plot estimated nonlinear effect of wind levels
tmp=fit$summary.random$level.wind[,c("mean","0.025quant","0.975quant")]
ylim=range(tmp)
plot(1:10,tmp$mean,type="l",ylim=ylim,lwd=2)
lines(1:10,tmp$`0.025quant`,col="blue",lwd=2)
lines(1:10,tmp$`0.975quant`,col="blue",lwd=2)

#Plot prediction for missing values (blue) and fitted values (grey):
xvals=1:nrow(airquality)
plot(xvals,airquality$Ozone,type="l",xlab="days",ylab="ozone",lwd=2,ylim=c(0,200))
idx.missing=which(is.na(airquality$Ozone))
mycolor=ifelse(is.na(airquality$Ozone),"blue","gray70")
mycex=ifelse(is.na(airquality$Ozone),1,.5)
points(xvals,fit$summary.fitted.values$mean,col=mycolor,pch=19,cex=mycex)



#Model 2c: Linear covariate effects and first-order auto-regression. ####
#There could be still be some dependence in ozone between consecutive days (after having taken account of wind and temperature, i.e. when looking at residuals of Model 1).
#We try to capture this with a first-order autoregressive structure over the observation days.

mydata=data.frame(y=airquality$Ozone,
                  intercept=1,
                  id.day=1:nrow(airquality),
                  temp=scale(airquality$Temp),
                  wind=scale(airquality$Wind)
)

myformula=y~-1+intercept+temp+wind+f(id.day,model="ar1")

#We fix the precision of the Gaussian likelihood to a very high value to avoid  identifiability problems between the AR1 precision and the precision of Gaussian likelihood.
#(When there is very weak dependence in the AR1 model, then both the AR1 and the Gaussian likelihood behave as "independent" effects, such that identifiability becomes a problem.)
fit=inla(myformula,
         family="gaussian",
         control.family=list(list(hyper=list(theta=list(initial=log(10^4),fixed=TRUE)))),
         data=mydata,
         control.fixed=list(prec=.0001),
         control.predictor=list(compute=TRUE))
fit2c=fit
summary(fit)
#Fixed effects: intercept, temperature and wind are significant as before.
#The estimate of the innovation standard deviation is (approximately):
sd.ar=sqrt(1/fit$summary.hyperpar$mean[1])
sd.ar #quite high...
resid=airquality$Ozone-fit$summary.fitted.values$mean
#histogram of residuals (probability histogram):
hist(resid,freq=F) #OK 
#(AR1 allows us to fit observations more or less exactly since we have 1 latent variable for reach observation)
#(But still, prediction "out-of-sample" may be very bad...)

#Plot prediction for missing values (blue) and fitted values (grey):
xvals=1:nrow(airquality)
plot(xvals,airquality$Ozone,type="l",xlab="days",ylab="ozone",lwd=2,ylim=c(0,200))
idx.missing=which(is.na(airquality$Ozone))
mycolor=ifelse(is.na(airquality$Ozone),"blue","gray70")
mycex=ifelse(is.na(airquality$Ozone),1,.5)
points(xvals,fit$summary.fitted.values$mean,col=mycolor,pch=19,cex=mycex)


#Can we compare models?
#Let's look at marginal likelihood pi(y):
fit2a$mlik
fit2b$mlik
fit2c$mlik
#=> 2b is best in terms of marginal likelihood.


#run a 29-fold cross-validation study to compare predictive performance:
n.notNA=sum(!is.na(airquality$Ozone))
n.notNA/29 #=> leave-4-out approach
#=> can create 29 subsamples of length 4

#permute indices with observed ozone values for cross-validation:
set.seed(1)
ids.for.cv=sample(which(!is.na(airquality$Ozone)),n.notNA)

pred.mod2a=pred.mod2b=pred.mod2c=c()

mydata=data.frame(y.cv=airquality$Ozone,
                  intercept=1,
                  id.day=1:nrow(airquality),
                  temp=scale(airquality$Temp),
                  wind=scale(airquality$Wind),
                  level.temp=level.temp,
                  level.wind=level.wind
) 

for(i in 1:29){
  cat("estimating cross-validation for hold-out dataset ", i, "\n")
  ids2pred=ids.for.cv[1:4+(i-1)*4] #which observations are held out and must be predicted?
  y.cv=airquality$Ozone
  y.cv[ids2pred]=NA #set data to NA for cross-validation
  mydata$y.cv=y.cv
  
  #Cross-validate hold-out dataset i for Model 2a:
  myformula=y.cv~-1+intercept+temp+wind
  fit=inla(myformula,
           family="gaussian",
           data=mydata,
           control.predictor=list(compute=TRUE),
           control.fixed=list(prec=.0001),
           control.inla=list(int.strategy="eb",strategy="gaussian")
           )
  pred.mod2a[ids2pred]=fit$summary.fitted.values$mean[ids2pred]
  
  #Cross-validate hold-out dataset i for Model 2b:
  myformula=y.cv~-1+intercept+f(level.temp,model="rw1",scale.model=TRUE,hyper=hyper.rw)+f(level.wind,scale.model=TRUE,model="rw1",hyper=hyper.rw)
  fit=inla(myformula,
           family="gaussian",
           data=mydata,
           control.predictor=list(compute=TRUE),
           control.fixed=list(prec=.0001),
           control.inla=list(int.strategy="eb",strategy="gaussian")
           )
  pred.mod2b[ids2pred]=fit$summary.fitted.values$mean[ids2pred]
  
  #Cross-validate hold-out dataset i for Model 2c:
  myformula=y.cv~-1+intercept+temp+wind+f(id.day,model="ar1")
  fit=inla(myformula,
           family="gaussian",
           control.family=list(hyper=list(theta=list(initial=log(10^4),fixed=TRUE))),
           data=mydata,
           control.predictor=list(compute=TRUE),
           control.inla=list(int.strategy="eb",strategy="gaussian"),
           control.fixed=list(prec=.0001)
           )
  #print(summary(fit))
  pred.mod2c[ids2pred]=fit$summary.fitted.values$mean[ids2pred]
}

#Here: look at root-mean squared error (note that there may be better predictive performance metrics than rmse):

myRMSE=function(observed,predicted){
  n=length(na.omit(observed))
  sqrt(1/n*sum((na.omit(observed)-na.omit(predicted))^2))
}

myRMSE(airquality$Ozone,pred.mod2a)
myRMSE(airquality$Ozone,pred.mod2b)
myRMSE(airquality$Ozone,pred.mod2c)
#Models are overall quite close.
#Model 2b has the best predictive performance. 

#Remark: As pointed out by one of the workshop participants (thanks Matthieu Authier), combining models 2 and 3 can lead to even better results. 


#Example 3: Curve smoothing. ####
#We first simulate a some noisy data with a highly nonlinear trend, and we then try to estimate this trend using a 1D SPDE model. 
#More information on the SPDE approach in Haakon's tutorial.

setwd("~/research/implementations/atelier-inla/INLA-first-steps")

#Generate some random data:
myfun=function(x){
  (x/2)^2+5*sin(x)
}

n=1000

set.seed(1)
x=sort(runif(n,0,10))
y=myfun(x)+rnorm(n,sd=3)
plot(x,y,type="l")
#dev.copy(pdf,file="dataspde1d.pdf",width=12)
#dev.off()

#define SPDE knots:
nknots=25
knots=quantile(x,probs=1:nknots/(nknots+1))

#fix left boundary to 0 here (assuming that we know that the trend function starts at 0):
mesh1d=inla.mesh.1d(loc=knots,interval=c(0,10),degree=2,boundary=c("dirichlet","free"))
idx.knots=1:mesh1d$n

#define Matérn model (with shape parameter 1, i.e. alpha=1.5=Matern shape + d/1 in the SPDE defined in dimension d=1) over SPDE knots with PC prior:
#here, we set prior probability of having range < 1 to 0.1, and of having standard deviation higher than 50 to 0.1.
#(these values are chosen approximately based on the "smoothness" and the "variability" that we can see in the plot of the data)
myspde=inla.spde2.pcmatern(mesh=mesh1d,alpha=1.5,prior.range=c(1,.1),prior.sigma=c(50,.1))

#Define the observation matrix A, which links the latent Gaussian variables (defined on mesh knots) to the response variable:
A=inla.spde.make.A(mesh1d,loc=x)

#data frame of other covariates (here only intercept)#
#we here define the intercept at the level of the 
covar.df=data.frame(intercept=rep(1,n))

#put everything into stack
#we here have 
mystack=inla.stack(data=list(y=y),A=list(A),effects=list(list(idx.knots=idx.knots)))

#INLA model formula:
#no need for an intercept here (we add -1) since we assume that the curve starts at 0.
form=y~-1+f(idx.knots,model=myspde)

#run INLA
numthr=2 #fix number of threads to 2 (since INLA may be quite greedy for CPU cores and memory)
#(This is not really an issue here since the data is small, but it can be advisable with higher-dimensional data and more complex models.)
fit=inla(form, 
         data=inla.stack.data(mystack), 
         family="gaussian",
         control.predictor=list(compute=FALSE,A=inla.stack.A(mystack)),
         num.threads=numthr,
         verbose=TRUE)
summary(fit)
#a summary of the latent Gaussian variables in the SPDE model is given here:
fit$summary.random$idx.knots
#(but this is not very easy to interpet...let's make a plot instead.)

#plot estimated curve and credible intervals on regular grid
n.grid=100
grid=inla.mesh.projector(mesh1d,dims=n.grid)
tmp=fit$summary.random$idx.knots[,c("mean","0.025quant","0.975quant")]
ylim=range(tmp,y)
par(mar=rep(1,4),oma=rep(1,4))
plot(x,y,cex=.5,col="gray50",pch=19,ylim=ylim)
lines(grid$x,inla.mesh.project(grid,field=tmp$mean),xlab="x",ylab="y",type="l",lwd=3,ylim=ylim)
lines(grid$x,inla.mesh.project(grid,field=tmp$`0.025quant`),col="blue",lwd=3)
lines(grid$x,inla.mesh.project(grid,field=tmp$`0.975quant`),col="blue",lwd=3)
#dev.copy(pdf,file="fitspde1d.pdf",width=12)
#dev.off()
