library(fields) #for plotting (image.plot)
library(INLA)
library(sp) #for transforming from latitude-longitude do Lambert 93 coordinates

#We set the working directory (MODIFY THIS):
setwd("~/research/implementations/atelier-inla/air-pollution/")


#General comment:
#This tutorial focuses on running INLA, not on producing fancy illustrations.

#We load the data objects for 2014 daily pollution data ###

#not used in the following code: CHIMERE predictions (see Maxime's talk)
#data frame with gridded hindcasts based on physical-chemical transport model (CHIMERE)
#load("CHIM_2014.RData")
##see how the data frame is structured:
#head(CHIM_daily2014)

#data frame with observations of daily PM10 and PM2.5 air pollutant concentrations at measurement stations ####
load("OBS_2014.RData")
#see how the data frame is structured:
head(OBS_daily2014)

#We keep only background stations (not traffic/industry, which generate highly localized pollution patterns):
OBS_daily2014=OBS_daily2014[OBS_daily2014$type_of_station=="Background",]
dim(OBS_daily2014) #more than 100,000 data points


dates=as.Date(OBS_daily2014$date)
OBS_daily2014$day=1+as.integer(difftime(OBS_daily2014$date,OBS_daily2014$date[1],units='days'))

#First, we transform longitude-latitude to Lambert93-coordinates (which correspond to "real" distances).
#This is necessary for accurate modeling of the spatial dependence since the distance of one unit latitude is different from one unit longitude.
#We use the EPSG codes for projections (European Petroleum Survey Group):
#4326 for longitude-latitude, 2154 for Lambert93
#For the spatial unit, we use 1000km. This is useful for the SPDE functions in RINLA, which may have stability problems with very large coordinate values.

xy_OBS=SpatialPoints(coords=OBS_daily2014[,c("long","lat")],proj4string=CRS("+init=epsg:4326"))
xy_OBS=spTransform(xy_OBS,CRS("+init=epsg:2154"))@coords/10^6

#We prepare the boundary coordinates of the French territory:
fr=map("world",regions="france")
#Transform coordinates to Lambert 93 using spTransform:
xy_fr = SpatialPoints(coords=na.omit(cbind(fr$x,fr$y)),proj4string=CRS("+init=epsg:4326"))
xy_fr = spTransform(xy_fr,CRS("+init=epsg:2154"))@coords/10^6
fr$x[!is.na(fr$x)] = xy_fr[,1]
fr$y[!is.na(fr$y)] = xy_fr[,2]
#Check:
map(fr) #looks OK
                            
#We plot observation sites and French territorial boundary:
plot(unique(xy_OBS),pch=19,cex=.33, xlab="x (1000 km)", ylab="y (1000 km)")
lines(fr$x,fr$y,lwd=2)


#We here use PM10 as our response variable. 
#(later, we could also consider PM2.5, e.g. for bivariate modeling of PM10 and PM2.5)


#Model 1a: Let's start with a simple model for spatial trends:  ####
#We use a Gaussian SPDE field as prior for the trend surface.
#Remark (based on the discussion during the workshop): The choice of prior distribution for the spatial field is not easy here since it seems that even nearby observation stations have very different trends sometimes. Moreover, the model disregards spatial and temporal dependence, i.e., it is rather simplistic.
#We further include x and y as linear covariates ("fixed effects"):

# First, we have to create a mesh (=triangulation of space).
#The Gaussian variables will be defined on the mesh nodes and are then interpolated linearly between the nodes.
#INLA:::inla.dynload.workaround() 
bound=inla.nonconvex.hull(unique(xy_OBS),convex=-.05)
bound2=inla.nonconvex.hull(unique(xy_OBS),convex=-.2)
plot(rbind(bound2$loc,bound2$loc[1,]),type="l")
lines(rbind(bound$loc,bound$loc[1,]))
points(unique(xy_OBS),pch=19,cex=.25)
#define minimum edge length as 10km and maximum edge length as 25km (within the study region) and 50km (in the extension zone).
mesh=inla.mesh.2d(loc=unique(xy_OBS),boundary=list(bound,bound2),cutoff=.01,max.edge=c(.025,.5),min.angle=c(21,21))
mesh$n
plot(mesh) #looks OK
lines(fr$x,fr$y,lwd=2,col="red") #add the true boundary of France in red

# definition for SPDE object with penalized complexity prior.
#Probability of having range < 10km is 10%, probability of having standard deviation larger than 25 is 50%
myspde=inla.spde2.pcmatern(mesh=mesh,alpha=2,prior.range=c(.01,.1),prior.sigma=c(25,0.5))

#observation matrix to make the connection between mesh nodes and observation sites:
A=inla.spde.make.A(mesh,loc=xy_OBS)

#vector of indices for the nodes of the mesh
idx.spatial=inla.spde.make.index("spatial",n.spde=mesh$n)

#create covariate dataframe:
#we add the intercept term 1, since in RINLA it's more comfortable to include the intercept explicitly.
covar.df=data.frame(intercept=1,x=xy_OBS[,1],y=xy_OBS[,2])

#create the stack with all the data and indices:
mystack=inla.stack(data=list(pm10=OBS_daily2014$PM10),A=list(A,1),effects=list(idx.spatial,covar.df))

#We write the formula for the model:
#(since we handle the intercept explicitly, we have to include "-1")
#Coordinates x and y are included as covariates.
#Recall that we have given the name "spatial" to the index of the spatial effect.
myformula=pm10~-1+intercept+x+y+f(spatial,model=myspde)

#Let's run inla:
#To avoid that inla becomes too greedy with memory and cpu, we may fix the maximum number of parallel threads to 2.
#For faster calculations, we here use control.inla=list(int.strategy="eb",strategy="gaussian").
#The default option would be control.inla=list(int.strategy="ccd",strategy="simplified.laplace")
#The following inla run should take approximatively 1 minute. 
num.thr=2 #number of threads that INLA is allowed to run in parallel
fit=inla(myformula,
         data=inla.stack.data(mystack),
         family="gaussian", #or: try "gamma", "sn"?
         control.predictor=list(A=inla.stack.A(mystack),compute=FALSE),
         control.inla=list(int.strategy="eb",strategy="gaussian"),
         verbose=FALSE,
         num.threads=num.thr
         )
summary(fit)
#very high standard errors for fixed effects...that's a bit weird.
#We should probably remove x and y which do not seem to contain useful information here:
#refit the model without fixed effects of coordinates x and y: 
myformula=pm10~-1+intercept+f(spatial,model=myspde)
fit=inla(myformula,
         data=inla.stack.data(mystack),
         family="gaussian", 
         control.predictor=list(A=inla.stack.A(mystack),compute=FALSE),
         control.inla=list(int.strategy="eb",strategy="gaussian"),
         verbose=FALSE,
         num.threads=num.thr
)
summary(fit)
#standard error for intercept looks more reasonable now...
#Range of spatial effect is approximately 20km (there will be very little smoothing over space in the estimated model.)
#Standard deviation of spatial effect is approximately 5.

#Plot the fitted trend: 
#Since standard plotting functions in R work on regular grids, we have to interpolate the estimated trend to such a grid, using the inla.mesh.project-mechanism:
proj_grid=inla.mesh.projector(mesh,xlim=range(bound$loc[,1]),ylim=range(bound$loc[,2]),dims=c(200,200))
#The estimated posterior mean of the spatial effect at the mesh values is as follows:
est.spatial.mesh=fit$summary.random$spatial$mean
#We use the projector proj_grid to interpolate the spatial effect values to a regular grid:
image.plot(proj_grid$x,proj_grid$y,inla.mesh.project(proj_grid,field=est.spatial.mesh))
points(unique(xy_OBS),pch=19,cex=.25)
#lines(rbind(bound$loc,bound$loc[1,]),lwd=2)
lines(fr$x,fr$y,lwd=2)

#The posterior standard deviation (=uncertainty) of the spatial effect at the mesh values is as follows:
sd.spatial.mesh=fit$summary.random$spatial$sd
#We use the projector proj_grid to interpolate the sd values to a regular grid:
image.plot(proj_grid$x,proj_grid$y,inla.mesh.project(proj_grid,field=sd.spatial.mesh))
points(unique(xy_OBS),pch=19,cex=.25)
#lines(rbind(bound$loc,bound$loc[1,]),lwd=2)
lines(fr$x,fr$y,lwd=2)
#=> Usually low uncertainty where we have many observations sites.
#Not very smooth overall...

#For making predictions of the trend over the full territory, we have to add the intercept to the latent spatial effect (here done in a "naive" way):
#Plot the trend surface:
image.plot(proj_grid$x,proj_grid$y,inla.mesh.project(proj_grid,field=est.spatial.mesh)+fit$summary.fixed$mean)
points(unique(xy_OBS),pch=19,cex=.25)
lines(fr$x,fr$y,lwd=2)
#lines(rbind(bound$loc,bound$loc[1,]),lwd=2)


#Alternatively, we could calculate the prediction on the grid in the very general way by adding NA data: ####
#observation matrix to make the connection between mesh nodes and observation sites:
proj_grid=inla.mesh.projector(mesh,xlim=range(bound$loc[,1]),ylim=range(bound$loc[,2]),dims=c(100,100))

#grid on wich we calculate the predictions:
xygrid=as.matrix(expand.grid(proj_grid$x,proj_grid$y))

#put prediction coordinates and observation coordinates into A:
A=inla.spde.make.A(mesh,loc=rbind(xygrid,xy_OBS))

#we add the intercept term 1, since in RINLA it's more comfortable to include the intercept explicitly.
covar.df=data.frame(intercept=1,x=c(xygrid[,1],xy_OBS[,1]),y=c(xygrid[,2],xy_OBS[,2]))

#create the stack with all the data and indices: add NA data here
mystack=inla.stack(data=list(pm10=c(rep(NA,nrow(xygrid)),OBS_daily2014$PM10)),A=list(A,1),effects=list(idx.spatial,covar.df),tag="mytag")

#When fitting, we set compute=TRUE (to calculate fitted and predicted values) and "link=1" (although only necessary when using non-identity link)
#NOTE: running INLA can now take considerably longer since we calculate the posterior estimation y.hat for each observation point y ("compute=TRUE")
fit=inla(myformula,
         data=inla.stack.data(mystack),
         family="gaussian",
         control.predictor=list(A=inla.stack.A(mystack),compute=TRUE,link=1),
         control.inla=list(int.strategy="eb",strategy="gaussian"),
         verbose=FALSE,
         num.threads=num.thr)
summary(fit)
#save(fit,file="fit1a.Rdata")
#load(file="fit1a.Rdata")

#We find the positions of the prediction values (there may be several ways to do this...):
idx.pred=inla.stack.index(mystack,tag="mytag")$data[1:nrow(xygrid)]
#We extract predictions:
pred.grid=fit$summary.fitted.values$mean[idx.pred]

#We plot the gridded posterior means:
image.plot(proj_grid$x,proj_grid$y,matrix(pred.grid,ncol=100,nrow=100))
points(unique(xy_OBS),pch=19,cex=.25)
lines(fr$x,fr$y,lwd=2)

#We extract and plot standard erors of predictions:
pred.grid=fit$summary.fitted.values$sd[idx.pred]
image.plot(proj_grid$x,proj_grid$y,matrix(pred.grid,ncol=100,nrow=100))
points(unique(xy_OBS),pch=19,cex=.25)
lines(fr$x,fr$y,lwd=2)

#We check the residuals (here: data minus prediction) ####
n.data=nrow(OBS_daily2014) #number of data points
idx.data=inla.stack.index(mystack,tag="mytag")$data[(nrow(xygrid)+1):(nrow(xygrid)+n.data)]
resid=OBS_daily2014$PM10-fit$summary.fitted.values$mean[idx.data]
hist(resid,breaks=100,freq=F)
#We add the density of the fitted Gaussian likelihood to the plot:
#therefore, we first calculate the posterior mean estimate of the Gaussian standard deviation
#(based on the posterior density of the Gaussian precision):
prec2sd=function(prec){#function to transform precision to standard deviation
  sqrt(1/prec)
}
sd.likelihood=inla.emarginal(prec2sd,marginal=fit$marginals.hyperpar$`Precision for the Gaussian observations`)
sd.likelihood
xvals=-500+1:1000
lines(xvals,dnorm(xvals,sd=sd.likelihood),col="blue",lwd=2)
#Conclusion: data are asymmetric and heavy-tailed (as opposed to our Gaussian response model).

#Alternatively, we could draw a QQ-plot:
qqplot(OBS_daily2014$PM10,fit$summary.fitted.values$mean[idx.data],xlab="observed",ylab="fitted",asp=1)
abline(0,1)
#This looks not so good, but is not surprising since the model is very simple.


#Model 1b: We add a new covariate given as the kernel estimate of the density of observations. ####
#This variable can be seen as a kind of proxy for population, hence traffic, hence pollution.
#What we do here is a simple approach to account for "preferential sampling" of the data. 

#First, we define a function to perform kernel smoothing of the spatial "intensity" of observation sites (using the Epanechnikov kernel):
mykernel = function(locations,obs,h){
  #DD2   = function(s1,s2) {z = (s1[1]-s2[1])^2 + (s1[2]-s2[2])^2} # Squared Eucl. Dist.
  K.EPA = function(dd2){ z = 0.75*sum( (1 - dd2)*(dd2 < 1) )} # Epanechnikov Kernels
  A = outer(locations[,1],obs[,1],FUN = "-")^2  + outer(locations[,2],obs[,2],FUN = "-")^2
  A = A / h^2
  Z = apply(A,MARGIN = 1,FUN=K.EPA) 
  return(Z)
}

#We calculate and plot the kernel estimate:
xy_station = unique(xy_OBS)
bw=70 #kernel bandwidth (in km)
OBSdens = matrix(mykernel(xygrid,xy_station,h=bw/1000),nrow=100,byrow=F)
x.plot = unique(xygrid[,1])
y.plot = unique(xygrid[,2])
image.plot(x.plot,y.plot,OBSdens) #unit is "number of sites per (100km)^2"
points(xy_station,pch=19,cex=.2,col="red")
map(fr,add=TRUE,lwd=2)

#We compute the "artificial" covariate given as the density of observation sites:
GRIDdens=mykernel(xygrid,xy_station,h=bw/1000)
OBSdens=mykernel(xy_OBS,xy_station,h=bw/1000)
covar.df=data.frame(intercept=1,x=c(xygrid[,1],xy_OBS[,1]),y=c(xygrid[,2],xy_OBS[,2]),dens=c(GRIDdens,OBSdens))

#We define the INLA formula with the density as covariate:
myformula=pm10~-1+intercept+dens+f(spatial,model=myspde)

#We create the stack with all the data and indices: add NA data here
A=inla.spde.make.A(mesh,loc=rbind(xygrid,xy_OBS))
mystack=inla.stack(data=list(pm10=c(rep(NA,nrow(xygrid)),OBS_daily2014$PM10)),A=list(A,1),effects=list(idx.spatial,covar.df),tag="mytag")

#When fitting, we set compute=TRUE (to calculate fitted and predicted values) and "link=1" (although only necessary when using non-identity link).
#NOTE: running INLA can now take considerably longer since we calculate the posterior estimation for each observation point y ("compute=TRUE"):
fit=inla(myformula,data=inla.stack.data(mystack),
         family="gaussian", #"sn"
         control.predictor=list(A=inla.stack.A(mystack),compute=TRUE,link=1),
         control.inla=list(int.strategy="eb",strategy="gaussian"),
         verbose=FALSE,
         num.threads=num.thr)
summary(fit) 
#We may save the fitted object for later use:
#save(fit,file="fit1b.Rdata")
#load(file="fit1b.Rdata")

#The density of measurement stations is significant here. 
#We will keep this covariate in later models!


#Model 1c: Same as Model 1b, but we use a log-gaussian model. ####
#We can simply change the link function to log:
#NOTE: Since this model is not gaussian, running takes much more time and memory here due to the Laplace approximation.
#We set compute=FALSE to avoid the long calculation of fitted values. 
#We also set a moderately informative PC prior on the precision. 
#(can take up to 10 minutes to run).
fit=inla(myformula,
        data=inla.stack.data(mystack),
        family="gaussian",
        control.predictor=list(A=inla.stack.A(mystack),compute=TRUE,link=1),
        control.family=list(list(control.link=list(model="log"),hyper=list(prec=list(initial=1,prior="pc.prec",param=c(5,.01))))),
        control.inla=list(int.strategy="eb",strategy="gaussian"),
        verbose=TRUE,
        num.threads=num.thr)
#save(fit,file="fit1c.Rdata")
#load(file="fit1c.Rdata")
summary(fit)
#If we compare the marginal likelihood pi(y) with preceding models (see summary(fit)),
# the results do not seem to have improved...
#We plot the histogram of residuals (on Gaussian scale)
resid=log(OBS_daily2014$PM10)-log(fit$summary.fitted.values$mean[idx.data])
hist(resid,breaks=100,freq=F) #looks symmetric
#What is the estimated Gaussian standard deviation in the response?
sd.likelihood=inla.emarginal(prec2sd,marginal=fit$marginals.hyperpar$`Precision for the Gaussian observations`)
sd.likelihood #Looks quite large as compared to residuals...not sure why???


#We could also try another likelihood with heavier-than-gaussian tails?
#names(inla.models()$likelihood)
#Maybe the gamma likelihood?
#(this is left as an exercice...)

#Model 2: Space-time model with daily SPDE models linked together through first-order autoregression####
#This model is quite complex with many latent Gaussian variables. 
#To keep estimation feasible on simple computers, we use only few data (10 days).
#In any case, running INLA will take some time (minutes to hours depending on the parameters...).
#We construct a model that allows us to produce a gridded prediction for the day 11, i.e. the day after the last day with observed data. 
ndays=10
sel.day=(31+28+1):(31+28+ndays)
idx.sel=which(OBS_daily2014$day %in% sel.day)
OBS_sel=OBS_daily2014[idx.sel,]
#We check if all days have data:
ndays == length(unique(OBS_sel$day)) #yes
#In our model, each day is a group of data, and we have to define a group index with values 1,2,...,10:
idgroup=OBS_sel$day-min(OBS_sel$day)+1
#How many observations for each day?:
table(idgroup) 

#We use a relatively coarse mesh here to speed up computations:
#mesh=inla.mesh.2d(loc=unique(xy_OBS[idx.sel,]),boundary=list(bound,bound2),cutoff=.01,max.edge=c(.08,1),min.angle=c(21,21))
mesh=inla.mesh.2d(loc=unique(xy_OBS[idx.sel,]),boundary=list(bound,bound2),cutoff=.05,max.edge=c(.2,1),min.angle=c(21,21))
mesh$n
plot(mesh) #looks OK
lines(fr$x,fr$y,lwd=2,col="red")

#We use the penalized complexity prior for the Matérn covariance: 
#Probability of having range < 20km is 5%, probability of having standard deviation larger than 25 is .1
myspde=inla.spde2.pcmatern(mesh=mesh,alpha=2,prior.range=c(.02,.05),prior.sigma=c(25,0.1))

#To predict on a grid for day 11, we define the grid.
#There are more or less elegant ways to do such prediction...
proj_grid=inla.mesh.projector(mesh,xlim=range(bound$loc[,1]),ylim=range(bound$loc[,2]),dims=c(50,50))
ngrid=100 #grid resolution
xgrid=min(bound$loc[,1])+(0.5+0:(ngrid-1))/ngrid*(max(bound$loc[,1])-min(bound$loc[,1]))
ygrid=min(bound$loc[,2])+(0.5+0:(ngrid-1))/ngrid*(max(bound$loc[,2])-min(bound$loc[,2]))
loc2pred=as.matrix(expand.grid(xgrid,ygrid))
LOC2PREDdens=mykernel(loc2pred,xy_station,h=bw/1000)
OBSdens=mykernel(xy_OBS[idx.sel,],xy_station,h=bw/1000)

#We create the observation matrix that links the latent variables on the mesh nodes to the response variables:
A=inla.spde.make.A(mesh,loc=rbind(xy_OBS[idx.sel,],loc2pred),group=c(idgroup,rep(ndays+1,ngrid^2)))

#We automatically create the vector of indices for the nodes of the mesh
idx.spatial=inla.spde.make.index("spatial",n.spde=mesh$n,n.group=ndays+1)

#We create the covariate dataframe:
#(we add the intercept term 1, since in RINLA it's more comfortable to include the intercept explicitly)
covar.df=data.frame(intercept=1,x=c(xy_OBS[idx.sel,1],loc2pred[,1]), y=c(xy_OBS[idx.sel,2],loc2pred[,2]),dens=c(LOC2PREDdens,OBSdens))

#We create the stack with all the data and indices:
mystack=inla.stack(data=list(pm10=c(OBS_daily2014$PM10[idx.sel],rep(NA,ngrid^2))),A=list(A,1),effects=list(idx.spatial,covar.df))

#We write the formula for the model:
#(since we handle the intercept explicitly, we have to include "-1")
#To simplify the model, we do not include the coordinates x and y.
#Recall that we have given the name "spatial" to the index of the spatial effect:
myformula=pm10~-1+intercept+dens+f(spatial,model=myspde,group=spatial.group,control.group=list(model="ar1"))

#run inla:
#To avoid that inla becomes too greedy with memory and cpu, we may fix the maximum number of parallel threads to 2.
#For faster calculations, we here use control.inla=list(int.strategy="eb",strategy="gaussian").
#The default option would be control.inla=list(int.strategy="ccd",strategy="simplified.laplace").
#Moreover, we approximately center the prior distribution of the intercept around the empirical average of PM10 values.
#We add "link=1" in control.predictor to have the correct link function for NA responses (although not necessary here since the link is identity).
#The following code takes approximately 1 minute to run:
num.thr=2
fit=inla(myformula,
         data=inla.stack.data(mystack),
        family="gaussian",  # family="sn"
         control.predictor=list(A=inla.stack.A(mystack),compute=TRUE,link=1),
        # control.fixed=list(prec=.01),
         control.inla=list(int.strategy="eb",strategy="gaussian"),
         num.threads=num.thr,
         verbose=TRUE)
summary(fit)
#save(fit,file="fit3.Rdata")
#load(file="fit3.Rdata")
#save(fit,file="fit3sn.Rdata")
#load(file=paste0("fit3sn.Rdata")

#We check the residuals:
idx.data=1:nrow(OBS_sel)
qqplot(OBS_sel$PM10,fit$summary.fitted.values$mean[idx.data],xlab="observed",ylab="fitted",asp=1)
abline(0,1)
#This looks rather good (although some problems with heavy tails in the data.)


#Plot the fitted model for day 11 (prediction):

#we first have to find the lines of the predicted values in the fit$summary.fitted.values object:
idx.pred=mystack$data$index[[1]][(nrow(OBS_sel)+1):(nrow(OBS_sel)+ngrid^2)]

image.plot(xgrid,ygrid,matrix(fit$summary.fitted.values$mean[idx.pred],nrow=ngrid,ncol=ngrid))
lines(fr$x,fr$y,lwd=2)
#add an illustration of the spatial variation of truly observed values at prediction day:
#our prediction day is:
day2pred=(31+28+ndays)+1
idx2pred=which(OBS_daily2014$day==day2pred)
mycex=0.5*OBS_daily2014[idx2pred,"PM10"]/mean(OBS_daily2014[idx2pred,"PM10"],na.rm=TRUE)
points(xy_OBS[idx2pred,],cex=mycex)

#We could also calculate the predictions at the observation sites of day 11, and the draw a QQ-plot to see how good they are (left as an exercice).


#We add an illustration of the spatial variation of observed values at the day before the prediction day
#(those values should have strong influence on the prediction):
image.plot(xgrid,ygrid,matrix(fit$summary.fitted.values$mean[idx.pred],nrow=ngrid,ncol=ngrid))
lines(fr$x,fr$y,lwd=2)
daybeforepred=(31+28+ndays)
idxbeforepred=which(OBS_daily2014$day==daybeforepred)
mycex=0.5*OBS_daily2014[idxbeforepred,"PM10"]/mean(OBS_daily2014[idxbeforepred,"PM10"],na.rm=TRUE)
points(xy_OBS[idxbeforepred,],cex=mycex)


#Question: Is the variable "dens" really useful for prediction? ####
#We compare the predictions of Model 2 for day 11  with and without "dens":

#We first refit the model with "dens" for prediction at location sites of day 11:####
idx.pred=which(OBS_daily2014$day %in% c(max(sel.day)+1))
loc2pred=xy_OBS[idx.pred,]
LOC2PREDdens=mykernel(loc2pred,xy_station,h=bw/1000)
OBSdens=mykernel(xy_OBS[idx.sel,],xy_station,h=bw/1000)

#We create the observation matrix that links the latent variables on the mesh nodes to the response variables:
A=inla.spde.make.A(mesh,loc=rbind(xy_OBS[idx.sel,],loc2pred),group=c(idgroup,rep(ndays+1,length(idx.pred))))

#We create the covariate dataframe:
#(we add the intercept term 1, since in RINLA it's more comfortable to include the intercept explicitly)
covar.df=data.frame(intercept=1,x=c(xy_OBS[idx.sel,1],loc2pred[,1]), y=c(xy_OBS[idx.sel,2],loc2pred[,2]),dens=c(LOC2PREDdens,OBSdens))

#We create the stack with all the data and indices:
mystack=inla.stack(data=list(pm10=c(OBS_daily2014$PM10[idx.sel],rep(NA,length(idx.pred)))),A=list(A,1),effects=list(idx.spatial,covar.df))

#We write the formula for the model:
myformula=pm10~-1+intercept+dens+f(spatial,model=myspde,group=spatial.group,control.group=list(model="ar1"))

#run inla:
num.thr=2
fit=inla(myformula,
         data=inla.stack.data(mystack),
         family="gaussian",  # family="sn"
         control.predictor=list(A=inla.stack.A(mystack),compute=TRUE,link=1),
         # control.fixed=list(prec=.01),
         control.inla=list(int.strategy="eb",strategy="gaussian"),
         num.threads=num.thr,
         verbose=TRUE)

idx.obs2pred=mystack$data$index[[1]][(nrow(OBS_sel)+1):(nrow(OBS_sel)+length(idx.pred))]
pred1=fit$summary.fitted.values$mean[idx.obs2pred]

myformula=pm10~-1+intercept+dens+f(spatial,model=myspde,group=spatial.group,control.group=list(model="ar1"))

#We now refit the model without "dens" for prediction at location sites of day 11:####

myformula=pm10~-1+intercept+f(spatial,model=myspde,group=spatial.group,control.group=list(model="ar1"))

#run inla:
num.thr=2
fit=inla(myformula,
         data=inla.stack.data(mystack),
         family="gaussian",  # family="sn"
         control.predictor=list(A=inla.stack.A(mystack),compute=TRUE,link=1),
         # control.fixed=list(prec=.01),
         control.inla=list(int.strategy="eb",strategy="gaussian"),
         num.threads=num.thr,
         verbose=TRUE)

idx.obs2pred=mystack$data$index[[1]][(nrow(OBS_sel)+1):(nrow(OBS_sel)+length(idx.pred))]
pred2=fit$summary.fitted.values$mean[idx.obs2pred]

#compare actual observation of day 11 with predictions of the two models:#
obs2pred=OBS_daily2014$PM10[idx.pred]

myRMSE=function(observed,predicted){
  n=length(na.omit(observed))
  sqrt(1/n*sum((na.omit(observed)-na.omit(predicted))^2))
}

isna=is.na(obs2pred)
rmse1=myRMSE(obs2pred[!isna],pred1[!isna])
rmse2=myRMSE(obs2pred[!isna],pred2[!isna])
rmse1
rmse2

#Remark: In the above fits, we have performed prediction by adding NA values in the response.
#An alternative would have been to create separate stacks for observed data and prediction data (NA), to combine those stacks, and then to plot the results pertaining to the prediction stack.


#EXERCICES:####
#1) Try to fit a skew-gaussian response distribution (family="sn" or family="sn2") or some other distribution instead of a Gaussian one, and compare. 
#2) Compare models based on criteria such as marginal likelihood, WAIC/DIC, PIT values.
#(need to add control.compute=list(dic=TRUE,waic=TRUE) in inla(...))
#Then you find the values in fit$waic$... and fit$dic$...
#4) Implement a full cross-validation scheme to compare the predictive performance of models (that's rather complicated and takes a long time to run...).
#5) Replace RMSE by an implementation Continuous ranked probability score where you can use the full posterior predictive distributions given in fit$marginals.fitted.values$...
#6) Instead of directly creating a single SPDE stack with oberved responses and NA responses for prediction, create two separate stacks (one for observations, one for predictions) and then "join them" using inla.stack.join(...). See Section 2.5 of the online SPDE book for an example. 
