library(fields)
library(INLA)
library(maptools)
library(sp)
library(spatstat)

#set your working directory:
setwd("~/research/implementations/atelier-inla/gorillas/")

#We use data of gorilla nesting site data collected in Cameroon:
#They contain spatial positions of gorilla nests, and several covariables are available.
#We will a Poisson response for the nest count in the cells of fine spatial grid. 
?gorillas
#plot gorilla nest positions (with information on group size and season):
plot(gorillas) 
#only (x,y)-coordinates:
plot(gorillas$x,gorillas$y,xlab="x",ylab="y",pch=19,cex=.5,asp=1)
#gridded covariates:
plot(gorillas.extra)

#For fitting the model, we will here use the grid of the covariates.
#Therefore, we transform the nest positions into nest counts for each of the grid cells:
Xim=as.im(gorillas,dimyx=gorillas.extra[[1]]$dim,W=window(gorillas))
Xim$dim
#We extract the range of  x- and y-coordinates for easier plotting later on.
xran=Xim$xrange
yran=Xim$yrange

#We extract the counts (=response variable in inla) for the pixels:
ycount=as.numeric(t(as.matrix(Xim)))
table(ycount) #we have several grid cells with more than one nest

#We prepare a dataframe with one column for each of the seven covariates:
covar.df=c()
for(i in 1:length(gorillas.extra)){
 covar.df=cbind(covar.df,as.numeric(t(as.matrix(gorillas.extra[[i]]))))
}
covar.df=as.data.frame(covar.df)
names(covar.df)=names(gorillas.extra)
xy.nests=cbind(gorillas$x,gorillas$y)
#We add the grid cell centers (x,y) to the covariates:
tmp=as.matrix(expand.grid(Xim$xcol,Xim$yrow))
covar.df$xcoor=tmp[,1]
covar.df$ycoor=tmp[,2]
#We add a column for the intercept:
covar.df$intercept=1
#We find the indices of missing covariates or reponse values (outside the observation window):
idx.na=as.vector(attr(na.omit(cbind(covar.df,ycount)),"na.action"))
idx.toKeep=setdiff(1:length(ycount),idx.na)
covar.df=covar.df[idx.toKeep,]
ycount=ycount[idx.toKeep]

#We rescale several numeric covariates to have mean 0 and variance 1: 
#(which can stabilize estimation in certain cases, and may simplify interpreting regression coefficients)
covar.df[,c(2,4,7)]=apply(covar.df[,c(2,4,7)],2,scale)

#Model 1: numerical covariates, and Aspect (=directional covariate) with 8 categories.####
#We here treat the 8 categories of Aspect as independent (we add "constr=TRUE" to avoid confusion with the intercept).
#The precision (=1/variance) of the Aspect coefficients is a hyperparameter that we estimate. 
#We define the INLA formula:
form=y~-1+intercept+elevation+slopeangle+waterdist+f(aspect,model="iid",constr=TRUE)
#(for simplicity, we here use default priors of random walk precisions and of iid precisions)
#We want to estimate the average number of nests per square km; therefore we add the multiplicative offset E=Xim$xstep*Xim$ystep/10^6 (the "E=..."-argument works only for the "poisson"-likelihood):
fit=inla(form,
         data=cbind(covar.df,y=ycount),
         family="poisson",
         E=Xim$xstep*Xim$ystep/10^6)
#(We can add "verbose=TRUE" in inla(...) if we want to have more technical information during the fitting process)
#We save the fitted object (for potential later use):
#save(fit,file="fit1.Rdata")
#We show a summary of the fit:
summary(fit)
# => All covariates except slopeangle are significant.
#By default, inla() calculates the posterior mean of the precision of the Aspect categories.
#The standard deviation may be easier to interpret.
#What would be the posterior mean of the standard deviation in this case?
#To do the transformation properly, we have to take into account that 
#"transformation of expected value" is different from "expected value of the transformation":
prec2sd=function(prec){#function to transform precision to standard deviation
  sqrt(1/prec)
}
sd.aspect=inla.emarginal(prec2sd,marginal=fit$marginals.hyperpar$`Precision for aspect`)
sd.aspect
#=> We have a standard deviation of approximately 0.3 in the coefficient estimates of Aspect categories. 
#We can compare to what we get by transforming "naively" the posterior mean of the precision:
sqrt(1/fit$summary.hyperpar$mean)

#We show summary of the fit of the aspect categories:
fit$summary.random$aspect

#We plot estimates and credible intervals for the categories of Aspect:
tmp=fit$summary.random$aspect[,c("ID", "mean", "0.025quant", "0.975quant")]
plot(tmp$ID,tmp$mean,pch=19,ylim=range(tmp[,2:4]),xlab="Aspect category", ylab="posterior estimate")
for(id in tmp$ID){
  lines(c(id,id),c(tmp$`0.025quant`[id],tmp$`0.975quant`[id]),col="blue",lwd=2)
}
lines(c(0,10^10),c(0,0),lwd=3,lty=2,col="gray50")
text(1:8,rep(-.6,8),levels(gorillas.extra$aspect))


#Model 2:  more complex model including all the categorical covariates, and a cyclic random walk for the Aspect covariate.####
#We define the INLA formula:
form=y~-1+intercept+elevation+slopeangle+waterdist+f(aspect,model="rw1",constr=TRUE,cyclic=TRUE)+f(heat,model="iid",constr=TRUE)+f(vegetation,model="iid",constr=TRUE)+f(slopetype,model="iid",constr=TRUE)
#(for simplicity, we here use default priors of random walk precisions and of iid precisions)
#We run INLA as before:
fit=inla(form,
         data=cbind(covar.df,y=ycount),
         family="poisson",
         E=Xim$xstep*Xim$ystep/10^6)
#save(fit,file="fit2.Rdata")
summary(fit)
#=> The marginal likelihood has improved quite a lot as compared to Model 1. 
#=> The differences in the estimated precisions of the various effects is quite strong.
#(therefore, it was good to estimate them, and not to fix them a priori)

#We plot estimates and credible intervals for the levels of vegetation:
fit$summary.random$vegetation
tmp=fit$summary.random$vegetation[,c("ID", "mean", "0.025quant", "0.975quant")]
plot(tmp$ID,tmp$mean,pch=19,ylim=range(tmp[,2:4]),xlab="vegation category", ylab="posterior estimate")
for(id in tmp$ID){
  lines(c(id,id),c(tmp$`0.025quant`[id],tmp$`0.975quant`[id]),col="blue",lwd=2)
}
lines(c(0,10^10),c(0,0),lwd=3,lty=2,col="gray50")

#We plot estimates and credible intervals for Aspect (here defined as a random walk):
#To stress the dependence between consecutive categories, we here plot the result with a continuous curve (although the true estimate is a step function!).
fit$summary.random$aspect
tmp=fit$summary.random$aspect[,c("ID", "mean", "0.025quant", "0.975quant")]
plot(tmp$ID,tmp$mean,pch=19,ylim=range(tmp[,2:4]),type="l",lwd=2,xlab="Aspect category", ylab="posterior estimate")
lines(tmp$ID,tmp$`0.025quant`,col="blue",lwd=2)
lines(tmp$ID,tmp$`0.975quant`,col="blue",lwd=2)
lines(c(0,10^10),c(0,0),lwd=3,lty=2,col="gray50")
# => The fit of Aspect looks quite similar to what we had in Model 1.

#Model 3: Could we have a nonlinear covariate effect in Elevation?####
#We here use a random walk over Elevation subdivided into 25 groups.
#Alternatively, we could have used a 1D SPDE model to do this. 
#INLA formula: 
form=y~-1+intercept+slopeangle+waterdist+
  f(aspect,model="rw1",constr=TRUE,cyclic=TRUE)+
  f(heat,model="iid",constr=TRUE)+f(vegetation,model="iid",constr=TRUE)+
  f(slopetype,model="iid",constr=TRUE)+
  f(inla.group(elevation,n=25),model="rw1")
#(for simplicity, we here use default priors of random walk precisions)
fit=inla(form,
         data=cbind(covar.df,y=ycount),
         family="poisson",
         E=Xim$xstep*Xim$ystep/10^6)
#save(fit,file="fit3.Rdata")
summary(fit)
#Remark: the marginal likelihood (mlik) increases slightly as compared to Model 2...

#We plot the nonlinear effect for elevation:
tmp=fit$summary.random$`inla.group(elevation, n = 25)`[,c("ID", "mean", "0.025quant", "0.975quant")]
ylim=range(tmp[,2:4])
#We had rescaled the Elevation covariate, we therefore "retransform" Elevation values to the original scale:
xvals=tmp$ID*sd(as.numeric(as.matrix(gorillas.extra$elevation)),na.rm=TRUE)+mean(as.numeric(as.matrix(gorillas.extra$elevation)),na.rm=TRUE)
plot(xvals,tmp$mean,pch=19,ylim=ylim,type="l",lwd=2,xlab="Elevation", ylab="posterior estimate")
lines(xvals,tmp$`0.025quant`,col="blue",lwd=2)
lines(xvals,tmp$`0.975quant`,col="blue",lwd=2)
lines(c(-10^10,10^10),c(0,0),lwd=3,lty=2,col="gray50")
# => maybe there is some slight nonlinearity...


#Model 4: we include a spatial effect to capture spatial variation not explained by the other covariates. ####
#To start simple, we create a model that contains only an intercept and the spatial effect.
#For the mesh used to define the spatial SPDE effect, we use a coarser discretization than the covariate grid (for computational reasons).
#We generate the mesh over an extended region such that we should have approximate stationarity within the observation window. 

bound=inla.nonconvex.hull(unique(cbind(covar.df$xcoor,covar.df$ycoor)),convex=-.05)
bound2=inla.nonconvex.hull(unique(cbind(covar.df$xcoor,covar.df$ycoor)),convex=-.25)
plot(rbind(bound2$loc,bound2$loc[1,]),type="l")
lines(rbind(bound$loc,bound$loc[1,]))
points(xy.nests,pch=19,cex=.25)

mesh=inla.mesh.2d(loc=cbind(gorillas$x,gorillas$y),boundary=list(bound,bound2),max.edge=c(400,1000),cutoff=100,min.angle=c(21,21))
mesh$n
plot(mesh)
points(gorillas$x,gorillas$y,cex=.5,pch=19,col="red")

#We define the SPDE model (with alpha=2 fixed) using the penalized complexity prior specification:
#Here, the probability of having range < 200m is 10%, probability of having standard deviation larger than 1 is .5:
spdemod=inla.spde2.pcmatern(mesh,prior.range=c(200,.1),prior.sigma=c(1,.5))
#We create the vector of indices for the nodes of the mesh:
idx.spatial=inla.spde.make.index("spatial",n.spde=spdemod$n.spde)
#We generate the observation matrix to make the connection between mesh nodes and observation sites:
A=inla.spde.make.A(mesh,loc=cbind(covar.df$xcoor,covar.df$ycoor))
dim(A)

#We create the stack with all the data and indices:
#(here we define the "intercept" covariate at the nodes of the mesh, i.e. within the SPDE model):
stack=inla.stack(data=list(y=ycount), A=list(A),effects=list(c(idx.spatial,list(intercept=1))))
#Alternatively, we could have added the "intercept" covariate at the level of the linear predictor (i.e., one value for each data point):
#stack=inla.stack(data=list(y=ycount), A=list(A,1),effects=list(idx.spatial,list(intercept=rep(1,nrow(covar.df)))))

#Formula: only intercept and spatial effect
form=y~-1+intercept+f(spatial,model=spdemod)
#We run inla:
#(in control.predictor, we have to indicate the observation matrix A establishing the link between the latent predictor variables and the response variable)
#(to get also the fitted values in the output, which takes much more computing time,  we have to set "control.predictor=list(A=inla.stack.A(stack),compute=TRUE)")
fit=inla(form,
         data=inla.stack.data(stack),
         family="poisson",
         control.predictor=list(A=inla.stack.A(stack)),
         E=Xim$xstep*Xim$ystep/10^6)
#save(fit,file="fit4.Rdata")
#With fitted values:
#fit=inla(form,
#         data=inla.stack.data(stack),
#         family="poisson",
#         control.predictor=list(A=inla.stack.A(stack),compute=TRUE),
#         E=Xim$xstep*Xim$ystep/10^6)
#save(fit,file="fit4b.Rdata") #with fitted values
summary(fit)
# => Dependence range is approximately 2km.
#(This is quite large, and it explains why there may be some "confounding" between the spatial effect and the intercept: the standard error is quite large for the intercept, and it is difficult to separate these two effects.)
# => Standard deviation of intercept (2) and of SPDE field are quite high (4).

#We plot the estimated latent spatial effect:
proj=inla.mesh.projector(mesh,xlim=xran,ylim=yran,dim=c(100,100))
est.grid=inla.mesh.project(proj,field=fit$summary.random$spatial$mean)
library(fields)
image.plot(x=proj$x,y=proj$y,z=est.grid)
points(gorillas$x,gorillas$y,cex=.2,pch=19,col="gray80")


#Model 5: an "all-inclusive" model with all covariate effects and spatial effect. ####
#To speed up computations, we fix most of the precision hyperparameters to the estimates from Model 2. 
#This model is quite complex and takes some time to estimate.
load(file="fit2.Rdata")
fit$summary.hyperpar #pour voir les valeurs estimées précédemment

#model formula:
form=y~-1+intercept+slopeangle+waterdist+elevation+
  f(aspect,model="rw1",constr=TRUE,cyclic=TRUE,hyper=list(prec=list(initial=log(40),fixed=TRUE)))+
  f(heat,model="iid",constr=TRUE,hyper=list(prec=list(initial=log(20000),fixed=TRUE)))+
  f(vegetation,model="iid",constr=TRUE,hyper=list(prec=list(initial=log(1),fixed=TRUE)))+
  f(slopetype,model="iid",constr=TRUE,hyper=list(prec=list(initial=log(20000),fixed=TRUE)))+f(spatial,model=spdemod)

#stack with all necessary data:
stack=inla.stack(data=list(y=ycount), A=list(A,1), effects=list(idx.spatial,covar.df))
#We run inla:
fit=inla(form,data=inla.stack.data(stack),
         family="poisson",
         control.predictor=list(A=inla.stack.A(stack)),
         E=Xim$xstep*Xim$ystep/10^6)
#save(fit,file="fit5.Rdata")
#We show the summary of the fit:
summary(fit)
#with fitted values (=predictions of the intensity):
#(very long to run, around 1 hour)
#fit=inla(form,data=inla.stack.data(stack),
#         family="poisson",
#         control.predictor=list(A=inla.stack.A(stack),compute=TRUE),
#         E=Xim$xstep*Xim$ystep/10^6)
##save(fit,file="fit5b.Rdata")


#Plot the spatial effect:
proj=inla.mesh.projector(mesh,xlim=xran,ylim=yran,dim=c(100,100))
est.grid=inla.mesh.project(proj,field=fit$summary.random$spatial$mean)
image.plot(x=proj$x,y=proj$y,z=est.grid)
points(gorillas$x,gorillas$y,cex=.2,pch=19,col="gray80")

#Plot the fitted values (only model fit5b with "compute=TRUE"):
#Since we do not have observations for all pixels of the rectangular grid, we here use quilt.plot() for plotting. 

load(file="fit5b.Rdata")

#We need the indices of the fitted values in fit$summary.fitted.values.
#Here it is relatively simple: 
idx.obs=1:length(ycount)
#Plot the fitted values (number of gorilla nests per 1km^2):
quilt.plot(covar.df$xcoor,covar.df$ycoor,fit$summary.fitted.values$mean[idx.obs],nx=128,ny=128)
points(xy.nests,col="gray80",pch=19,cex=.2)
#Plot on log-scale where spatial variability is smaller:
quilt.plot(covar.df$xcoor,covar.df$ycoor,log(fit$summary.fitted.values$mean[idx.obs]),nx=128,ny=128)
points(xy.nests,col="gray80",pch=19,cex=.2)


#How could we compare fitted models? How could we choose the best model? ####

#1) Marginal likelihood (probability of the data under the model)
#(will choose rather complex models and can lead to overfitting)
#2) Information criteria: DIC, WAIC? 
#(we have to include control.compute=list(...,dic=TRUE,waic=TRUE) in the inla call)
#They measure the goodness-of-fit while penalizing the complexity of the model to avoid overfitting.
#3) Predictive performance metrics based on cross-validation of data.
#We could calculate sensitivity/specificity-values for presence-absence, based on cross-validated predictions of the gorilla nest density => ROC curves, AUC values.
#Cross-validation would be quite computer-intensive here, but should be relatively simple to implement.
#4) ...?
